@@ -108,6 +108,7 @@ class Context:
         else:
             for symbol in symbols:
                 self.module_symbols[symbol] = module, symbols[symbol]
+        module.on_import()
 
     def declare_global(self, name, ast_node=None):
         """Declare a global variable.
@@ -317,12 +318,16 @@ class Module:
     """Module known to transpiler.
     """
 
-    def __init__(self, name, constants=None, variables=None, functions=None):
+    def __init__(self, transpiler, name, constants=None, variables=None, functions=None):
+        self.transpiler = transpiler
         self.name = name
         self.constants = constants or {}
         self.variables = variables or {}
         self.functions = functions or {}
 
+    def on_import(self):
+        pass
+
 
 class ATranspiler:
     """Transpiler from a subset of Python3 to Aseba.
@@ -353,12 +358,29 @@ class ATranspiler:
         # dict module_name: module (modules which can be imported)
         self.modules = {}
 
+        # onevent_preamble[E]: set of code fragments prepended to transpiled
+        # @onevent def E or definition of onevent E if no @onevent is given for E
+        self.onevent_preamble = {}
+        # associated variable declarations
+        self.additional_var_declarations = {}
+
     def set_preamble(self, preamble):
         """Set the Python source code compiled just before source
         (typically import statements).
         """
         self.preamble = preamble
 
+    def add_onevent_preamble(self, name, src_aseba, var_decl=None):
+        """Add source code to be prepended to transpiled @onevent definition.
+        """
+        if name in self.onevent_preamble:
+            self.onevent_preamble[name].add(src_aseba)
+        else:
+            self.onevent_preamble[name] = {src_aseba}
+        if var_decl is not None:
+            self.additional_var_declarations = {**self.additional_var_declarations,
+                                                **var_decl}
+
     def set_source(self, source):
         """Set the Python source code and reset transpilation.
         """
@@ -856,7 +878,7 @@ end
             if isinstance(expr, ast.Ellipsis):
                 return ""
             # hard-coded constants, such as strings used for documentation
-            if isinstance(expr, ast.Constant):
+            if isinstance(expr, (ast.Constant, ast.Str)):
                 return ""
             # special functions
             if isinstance(expr, ast.Call):
@@ -870,11 +892,14 @@ end
                     return aux_statements
                 if fun_name == "emit":
                     # hard-coded emit(name, params...) -> "emit name [params]"
-                    if (len(expr.args) < 1 or
-                        not isinstance(expr.args[0], ast.Constant) or
-                        not isinstance(expr.args[0].value, str)):
+                    event_name = None
+                    if len(expr.args) >= 1:
+                        if isinstance(expr.args[0], ast.Constant) and isinstance(expr.args[0].value, str):
+                            event_name = expr.args[0].value
+                        elif isinstance(expr.args[0], ast.Str):
+                            event_name = expr.args[0].s
+                    if event_name is None:
                         raise TranspilerError("bad event name in emit", node)
-                    event_name = expr.args[0].value
                     event_size = len(expr.args) - 1
                     if event_name in self.events_out:
                         if event_size != self.events_out[event_name]:
@@ -912,6 +937,8 @@ end
                     for i, arg in enumerate(expr.args):
                         if isinstance(arg, ast.Constant) and isinstance(arg.value, str):
                             print_format_string += (" " if i > 0 else "") + arg.value.replace("%", "%%")
+                        elif isinstance(arg, ast.Str):
+                            print_format_string += (" " if i > 0 else "") + arg.s.replace("%", "%%")
                         else:
                             value, aux_st, _ = self.compile_expr(arg, context, self.PRI_NUMERIC)
                             aux_statements += aux_st
@@ -1146,6 +1173,8 @@ return
                 function_src += f"""
 onevent {fun_name.replace("_", ".")}
 """
+                if fun_name in self.onevent_preamble:
+                    function_src += "".join(self.onevent_preamble[fun_name])
                 for i, arg in enumerate(function.function_def.args.args):
                     function_src += f"""{function.var_str(arg.arg, True)} = event.args[{i}]
 """
@@ -1154,6 +1183,13 @@ onevent {fun_name.replace("_", ".")}
 sub {fun_name}
 """
             function_src += fun_output_src
+        # onevent_preamble for functions not defined in Python source code
+        for fun_name in self.onevent_preamble:
+            if fun_name not in context_top.functions:
+                function_src += f"""
+onevent {fun_name.replace("_", ".")}
+"""
+                function_src += "".join(self.onevent_preamble[fun_name])
 
         # compile top-level code again, now that function return types are known
         self.output_src = self.compile_node_array(top_code, context_top) + function_src
@@ -1169,6 +1205,9 @@ sub {fun_name}
             context_top.functions[fun_name].var_declarations()
             for fun_name in context_top.functions
         ])
+        for var_name, var_val0 in enumerate(self.additional_var_declarations):
+            var_decl += f"""var {var_name} = {var_val0}
+"""
         if len(var_decl) > 0:
             self.output_src = var_decl + "\n" + self.output_src
 
@@ -1229,8 +1268,10 @@ sub {fun_name}
         transpiler = ATranspiler()
         if modules == "thymio":
             from tdmclient.module_thymio import ModuleThymio
+            from tdmclient.module_clock import ModuleClock
             modules = {
-                "thymio": ModuleThymio()
+                "thymio": ModuleThymio(transpiler),
+                "clock": ModuleClock(transpiler),
             }
         if modules is not None:
             transpiler.modules = {**transpiler.modules, **modules}
