@@ -13,11 +13,15 @@ import re
 from tdmclient import ClientAsync, ArrayCache
 from tdmclient.atranspiler import ATranspiler
 from tdmclient.module_thymio import ModuleThymio
+from tdmclient.module_clock import ModuleClock
 
 
 class TDMConsole(code.InteractiveConsole):
 
-    def __init__(self, local_var=None, define_functions=True):
+    def __init__(self,
+                 local_var=None,
+                 define_functions=True,
+                 user_functions=None):
         """New interactive console with synchronization with TDM node.
 
         Argument:
@@ -49,17 +53,13 @@ class TDMConsole(code.InteractiveConsole):
             """
 
             # send and flush all variables which might have been changed
-            if len(self.var_set) > 0:
-                for name in self.var_set:
-                    self.send_variable(name, self.local_var[name])
-                self.flush_variables()
+            self.send_variables(self.var_set)
 
             # wait
             ClientAsync.aw(self.client.sleep(t))
 
             # fetch all variables which might be used
-            for name in self.var_got:
-                self.local_var[name] = self.fetch_variable(name)
+            self.fetch_variables(self.var_got, node_flush=False)
 
         def robot_code(language="python"):
             """Gather Python or Aseba source code for the robot.
@@ -71,8 +71,8 @@ class TDMConsole(code.InteractiveConsole):
             src = ""
 
             # robot variables
-            for name in self.robot_var_set:
-                name_py = self.to_python_name(name)
+            for name_a in self.robot_var_set:
+                name_py = self.to_python_name(name_a)
                 src += f"""{name_py} = {self.local_var[name_py]}
 """
 
@@ -131,6 +131,24 @@ class TDMConsole(code.InteractiveConsole):
             """
             self.stop_program(discard_output=True)
 
+        def get_var(*args):
+            """Get robot variables passed as a set or list of names and
+            return them in a tuple.
+            """
+            self.fetch_variables(args)
+            return tuple(
+                self.local_var[name_py]
+                for name_py in args
+            )
+
+        def set_var(**kwargs):
+            """Set robot variables passed as keyword arguments.
+            """
+            for name_py in kwargs:
+                self.local_var[name_py] = kwargs[name_py]
+                self.send_variable(name_py, kwargs[name_py])
+            self.flush_variables()
+
         def clear_event_data(event_name=None):
             """Clear all or named event data.
             """
@@ -160,9 +178,12 @@ class TDMConsole(code.InteractiveConsole):
             "robot_code_new": robot_code_new,
             "run": run,
             "stop": stop,
+            "get_var": get_var,
+            "set_var": set_var,
             "clear_event_data": clear_event_data,
             "get_event_data": get_event_data,
             "send_event": send_event,
+            **(user_functions if user_functions is not None else {}),
         }
 
         if local_var is None:
@@ -177,6 +198,9 @@ class TDMConsole(code.InteractiveConsole):
             if define_functions:
                 self.local_var.update(self.functions)
 
+        # from initial node description
+        self.sync_var_vm = None
+        # completed by variable change messages
         self.sync_var = None
 
         # for generating source code for robot
@@ -200,14 +224,23 @@ class TDMConsole(code.InteractiveConsole):
         # fetch all variables
         await self.node.wait_for_variables()
         sync_var = set()
-        for name in self.node.vm_description["variables"]:
-            value = self.node[name]
+        for name_a in self.node.vm_description["variables"]:
+            value = self.node[name_a]
             if isinstance(value, ArrayCache):
                 value = list(value)
-            name_py = self.to_python_name(name)
+            name_py = self.to_python_name(name_a)
             self.local_var[name_py] = value
             sync_var.add(name_py)
-        self.sync_var = sync_var
+        self.sync_var_vm = sync_var
+        self.sync_var = sync_var.copy()
+
+        # complete self.sync_var when receiving values for unknown variables
+        def update_sync_var(node, variables):
+            for name in variables:
+                if variables[name] is not None:
+                    self.sync_var.add(name)
+
+        self.node.add_variables_changed_listener(update_sync_var)
 
     @staticmethod
     def transpile(src, import_thymio=True):
@@ -217,10 +250,11 @@ class TDMConsole(code.InteractiveConsole):
             src -- Python source code
             import_thymio -- if True (default), predefine all Thymio symbols
         """
-        modules = {
-            "thymio": ModuleThymio()
-        }
         transpiler = ATranspiler()
+        modules = {
+            "thymio": ModuleThymio(transpiler),
+            "clock": ModuleClock(transpiler),
+        }
         transpiler.modules = {**transpiler.modules, **modules}
         if import_thymio:
             transpiler.set_preamble("""from thymio import *
@@ -241,6 +275,52 @@ class TDMConsole(code.InteractiveConsole):
         else:
             return self.event_data_dict[event_name] if event_name in self.event_data_dict else []
 
+    def reset_sync_var(self):
+        self.sync_var = self.sync_var_vm.copy()
+
+    def process_events(self, on_event_data=None):
+        """Listen to events sent by the program running on the robot and process
+        them until _exit is received.
+
+        Argument:
+            on_event_data -- func(event_name) called when new data is received
+        """
+
+        exit_received = None  # or exit code once received
+
+        def on_event_received(node, event_name, event_data):
+            if self.output_enabled:
+                if event_name == "_exit":
+                    nonlocal exit_received
+                    exit_received = event_data[0]
+                elif event_name == "_print":
+                    print_id = event_data[0]
+                    print_format, print_num_args = print_statements[print_id]
+                    print_args = tuple(event_data[1 : 1 + print_num_args])
+                    print_str = print_format % print_args
+                    print(print_str)
+                else:
+                    if len(event_data) > 0:
+                        if event_name not in self.event_data_dict:
+                            self.event_data_dict[event_name] = []
+                        self.event_data_dict[event_name].append(event_data)
+                        if on_event_data is not None:
+                            on_event_data(event_name)
+
+        def wake():
+            return exit_received is not None
+
+        self.client.clear_event_received_listeners()
+        self.client.add_event_received_listener(on_event_received)
+        try:
+            ClientAsync.aw(self.node.watch(events=True))
+            ClientAsync.aw(self.client.sleep(wake=wake))
+            self.stop_program(discard_output=True)
+            if exit_received:
+                print(f"Exit, status={exit_received}")
+        finally:
+            self.client.clear_event_received_listeners()
+
     def run_program(self, src, language="aseba", wait=False, import_thymio=True):
         print_statements = []
         events = []
@@ -258,6 +338,7 @@ class TDMConsole(code.InteractiveConsole):
             for event_name in transpiler.events_out:
                 events.append((event_name, transpiler.events_out[event_name]))
             if len(events) > 0:
+                events = ClientAsync.aw(self.node.filter_out_vm_events(events))
                 ClientAsync.aw(self.node.register_events(events))
         elif language != "aseba":
             raise Exception(f"Unsupported language {language}")
@@ -265,6 +346,7 @@ class TDMConsole(code.InteractiveConsole):
         error = ClientAsync.aw(self.node.compile(src))
         if error is not None:
             raise Exception(error["error_msg"])
+        self.client.clear_event_received_listeners()
         exit_received = None  # or exit code once received
         if wait is None:
             # wait if there are events to receive
@@ -286,20 +368,23 @@ class TDMConsole(code.InteractiveConsole):
                             if event_name not in self.event_data_dict:
                                 self.event_data_dict[event_name] = []
                             self.event_data_dict[event_name].append(event_data)
-            self.client.clear_event_received_listeners()
             self.client.add_event_received_listener(on_event_received)
             ClientAsync.aw(self.node.watch(events=True))
+        self.reset_sync_var()
         error = ClientAsync.aw(self.node.run())
         if error is not None:
             raise Exception(f"Error {error['error_code']}")
         self.node.send_set_scratchpad(src)
         if wait:
-            def wake():
-                return exit_received is not None
-            ClientAsync.aw(self.client.sleep(wake=wake))
-            self.stop_program(discard_output=True)
-            if exit_received:
-                print(f"Exit, status={exit_received}")
+            try:
+                def wake():
+                    return exit_received is not None
+                ClientAsync.aw(self.client.sleep(wake=wake))
+                self.stop_program(discard_output=True)
+                if exit_received:
+                    print(f"Exit, status={exit_received}")
+            finally:
+                self.client.clear_event_received_listeners()
 
     def stop_program(self, discard_output=False):
         output_enabled_orig = self.output_enabled
@@ -311,31 +396,57 @@ class TDMConsole(code.InteractiveConsole):
         finally:
             self.output_enabled = output_enabled_orig
 
-    @staticmethod
-    def from_python_name(p_name):
-        # replace underscores with dots, except first underscore
-        return re.sub(r"(?<=.)_", r".", p_name)
+    def from_python_name(self, name_py):
+        # replace underscores with dots, except first underscore, for node variables
+        if self.node is not None:
+            name_a = re.sub(r"(?<=.)_", r".", name_py)
+            if name_a in self.node.var:
+                return name_a
+        # not a node variable: keep as is
+        return name_py
 
     @staticmethod
-    def to_python_name(a_name):
+    def to_python_name(name_a):
         # replace dots with underscores
-        return a_name.replace(".", "_")
+        return name_a.replace(".", "_")
 
-    def fetch_variable(self, name):
-        self.node.flush()
-        value = self.node[self.from_python_name(name)]
+    def fetch_variable(self, name_py, node_flush=True):
+        """Fetch a variable from the robot and return its value.
+        """
+
+        if node_flush:
+            self.node.flush()
+        value = self.node[self.from_python_name(name_py)]
         if isinstance(value, ArrayCache):
             value = list(value)
         return value
 
-    def send_variable(self, name, value):
-        a_name = self.from_python_name(name)
-        self.robot_var_set.add(a_name)
-        self.node[a_name] = value
+    def fetch_variables(self, names_py, node_flush=True):
+        """Fetch variables from the robot and store their values into
+        self.local_var.
+        """
+
+        if node_flush:
+            self.node.flush()
+        for name_py in names_py:
+            self.local_var[name_py] = self.fetch_variable(name_py,
+                                                          node_flush=False)
+
+    def send_variable(self, name_py, value):
+        name_a = self.from_python_name(name_py)
+        self.robot_var_set.add(name_a)
+        self.node[name_a] = value
 
     def flush_variables(self):
         self.node.flush()
 
+    def send_variables(self, names_py, node_flush=True):
+        if len(names_py) > 0:
+            for name_py in names_py:
+                self.send_variable(name_py, self.local_var[name_py])
+            if node_flush:
+                self.flush_variables()
+
     def find_global_var(self, nodes, globals=None):
         """Return variable referenced in expressions, variables assigned to,
         variables declared as global, and function called (tuple of 4 sets).
@@ -478,7 +589,7 @@ class TDMConsole(code.InteractiveConsole):
                              ast.Delete, ast.FunctionDef,
                              ast.Import, ast.ImportFrom,
                              ast.Lambda,
-                             ast.Nonlocal,
+                             ast.NameConstant, ast.Nonlocal, ast.Num,
                              ast.Pass)):
                 pass
             elif node is not None:
@@ -515,8 +626,7 @@ class TDMConsole(code.InteractiveConsole):
             self.var_got, self.var_set, _, _ = self.find_global_var(self.cmd_tree.body)
             self.var_got &= self.sync_var
             self.var_set &= self.sync_var
-            for name in self.var_got:
-                self.local_var[name] = self.fetch_variable(name)
+            self.fetch_variables(self.var_got)
         except Exception as e:
             # print("pre_run error", e)
             pass
@@ -558,10 +668,7 @@ class TDMConsole(code.InteractiveConsole):
         """Analyze a complete command after it has been executed,
         with or without error.
         """
-        if len(self.var_set) > 0:
-            for name in self.var_set:
-                self.send_variable(name, self.local_var[name])
-            self.flush_variables()
+        self.send_variables(self.var_set)
         try:
             if (self.cmd_tree is not None and
                 self.cmd_tree.body is not None):
