@@ -13,6 +13,7 @@ import re
 from tdmclient import ClientAsync
 from tdmclient.atranspiler import ATranspiler
 from tdmclient.module_thymio import ModuleThymio
+from tdmclient.module_clock import ModuleClock
 
 
 def help(**kwargs):
@@ -20,10 +21,12 @@ def help(**kwargs):
 Run program on robot, from file or stdin
 
 Options:
-  --debug n      display diagnostic information (0=none, 1=basic, 2=more, 3=verbose)
+  --debug=n      display diagnostic info (0=none, 1=basic, 2=more, 3=verbose)
+  --event=N      register custom event without data
+  --event=N[S]   register custom event with data of the specified size
   --help         display this help message and exit
   --language=L   programming language (aseba or python); default=automatic
-  --nosleep      exit immediately (default with neither events nor print statement)
+  --nosleep      exit immediately (default with no events, print() or exit())
   --nothymio     don't import the symbols of thymio library
   --robotid=I    robot id; default=any
   --robotname=N  robot name; default=any
@@ -162,10 +165,11 @@ if __name__ == "__main__":
 
     if language == "python":
         # transpile from Python to Aseba
-        modules = {
-            "thymio": ModuleThymio()
-        }
         transpiler = ATranspiler()
+        modules = {
+            "thymio": ModuleThymio(transpiler),
+            "clock": ModuleClock(transpiler),
+        }
         transpiler.modules = {**transpiler.modules, **modules}
         if import_thymio:
             transpiler.set_preamble("""from thymio import *
@@ -183,13 +187,10 @@ if __name__ == "__main__":
         for event_name in transpiler.events_out:
             events.append((event_name, transpiler.events_out[event_name]))
 
-    if sleep is None:
-        sleep = len(events) > 0
-
     with ClientAsync(tdm_addr=tdm_addr, tdm_port=tdm_port, debug=debug) as client:
 
         async def prog():
-            global status
+            global status, events, sleep
             with await client.lock(node_id=robot_id, node_name=robot_name) as node:
                 if stop:
                     error = await node.stop()
@@ -199,7 +200,10 @@ if __name__ == "__main__":
                 else:
                     if scratchpad < 2:
                         if len(events) > 0:
+                            events = await node.filter_out_vm_events(events)
                             await node.register_events(events)
+                            if sleep is None:
+                                sleep = len(events) > 0
                         error = await node.compile(program)
                         if error is not None:
                             print(f"Compilation error: {error['error_msg']}")
